@@ -1,5 +1,5 @@
 # --- BEGIN COPYRIGHT BLOCK ---
-# Copyright (C) 2015 Red Hat, Inc.
+# Copyright (C) 2020 Red Hat, Inc.
 # All rights reserved.
 #
 # License: GPL (version 3 or any later version).
@@ -11,7 +11,7 @@ import copy
 import ldap
 from lib389._constants import *
 from lib389.properties import *
-from lib389.utils import normalizeDN, ensure_str, ensure_bytes
+from lib389.utils import normalizeDN, ensure_str, assert_c
 from lib389 import Entry
 
 # Need to fix this ....
@@ -19,21 +19,21 @@ from lib389 import Entry
 from lib389._mapped_object import DSLdapObjects, DSLdapObject
 from lib389.mappingTree import MappingTrees
 from lib389.exceptions import NoSuchEntryError, InvalidArgumentError
-from lib389.replica import Replicas
+from lib389.replica import Replicas, Changelog
 from lib389.cos import (CosTemplates, CosIndirectDefinitions,
                         CosPointerDefinitions, CosClassicDefinitions)
 
 # We need to be a factor to the backend monitor
 from lib389.monitor import MonitorBackend
 from lib389.index import Index, Indexes, VLVSearches, VLVSearch
-from lib389.tasks import ImportTask, ExportTask, CleanAllRUVTask, Tasks
+from lib389.tasks import ImportTask, ExportTask, Tasks
 from lib389.encrypted_attributes import EncryptedAttr, EncryptedAttrs
 
 
 # This is for sample entry creation.
 from lib389.configurations import get_sample_entries
 
-from lib389.lint import DSBLE0001, DSBLE0002, DSBLE0003, DSVIRTLE0001
+from lib389.lint import DSBLE0001, DSBLE0002, DSBLE0003, DSVIRTLE0001, DSCLLE0001
 
 
 class BackendLegacy(object):
@@ -341,11 +341,11 @@ class BackendLegacy(object):
 
     def getProperties(self, suffix=None, backend_dn=None, bename=None,
                       properties=None):
-        raise NotImplemented
+        raise NotImplementedError
 
     def setProperties(self, suffix=None, backend_dn=None, bename=None,
                       properties=None):
-        raise NotImplemented
+        raise NotImplementedError
 
     def toSuffix(self, entry=None, name=None):
         '''
@@ -393,6 +393,10 @@ class BackendLegacy(object):
         replace = [(ldap.MOD_REPLACE, 'nsslapd-require-index', 'on')]
         self.modify_s(dn, replace)
 
+    @classmethod
+    def lint_uid(cls):
+        return 'backends'
+
 
 class Backend(DSLdapObject):
     """Backend DSLdapObject with:
@@ -413,19 +417,22 @@ class Backend(DSLdapObject):
         self._must_attributes = ['nsslapd-suffix', 'cn']
         self._create_objectclasses = ['top', 'extensibleObject', BACKEND_OBJECTCLASS_VALUE]
         self._protected = False
-        self._lint_functions = [self._lint_mappingtree, self._lint_search, self._lint_virt_attrs]
         # Check if a mapping tree for this suffix exists.
         self._mts = MappingTrees(self._instance)
 
+    def lint_uid(self):
+        return self.get_attr_val_utf8_l('cn').lower()
+
     def _lint_virt_attrs(self):
         """Check if any virtual attribute are incorrectly indexed"""
+        bename = self.lint_uid()
         indexes = self.get_indexes()
         suffix = self.get_attr_val_utf8('nsslapd-suffix')
-
         # First check nsrole
         try:
             indexes.get('nsrole')
             report = copy.deepcopy(DSVIRTLE0001)
+            report['check'] = f'backends:{bename}:virt_attrs'
             report['detail'] = report['detail'].replace('ATTR', 'nsrole')
             report['fix'] = report['fix'].replace('ATTR', 'nsrole')
             report['fix'] = report['fix'].replace('SUFFIX', suffix)
@@ -438,7 +445,7 @@ class Backend(DSLdapObject):
 
         # Check COS next
         for cosDefType in [CosIndirectDefinitions, CosPointerDefinitions, CosClassicDefinitions]:
-            defs = cosDefType(self._instance, self._dn).list()
+            defs = cosDefType(self._instance, suffix).list()
             for cosDef in defs:
                 attrs = cosDef.get_attr_val_utf8_l("cosAttribute").split()
                 for attr in attrs:
@@ -449,6 +456,7 @@ class Backend(DSLdapObject):
                         indexes.get(attr)
                         # If we got here there is an index (bad)
                         report = copy.deepcopy(DSVIRTLE0001)
+                        report['check'] = f'backends:{bename}:virt_attrs'
                         report['detail'] = report['detail'].replace('ATTR', attr)
                         report['fix'] = report['fix'].replace('ATTR', attr)
                         report['fix'] = report['fix'].replace('SUFFIX', suffix)
@@ -465,16 +473,19 @@ class Backend(DSLdapObject):
         """Perform a search and make sure an entry is accessible
         """
         dn = self.get_attr_val_utf8('nsslapd-suffix')
+        bename = self.lint_uid()
         suffix = DSLdapObject(self._instance, dn=dn)
         try:
             suffix.get_attr_val('objectclass')
         except ldap.NO_SUCH_OBJECT:
             # backend root entry not created yet
             DSBLE0003['items'] = [dn, ]
+            DSBLE0003['check'] = f'backends:{bename}:search'
             yield DSBLE0003
         except ldap.LDAPError as e:
             # Some other error
             DSBLE0002['detail'] = DSBLE0002['detail'].replace('ERROR', str(e))
+            DSBLE0002['check'] = f'backends:{bename}:search'
             DSBLE0002['items'] = [dn, ]
             yield DSBLE0002
 
@@ -485,19 +496,39 @@ class Backend(DSLdapObject):
         * missing mapping tree entries for the backend
         * missing indices if we are local and have log access?
         """
-
         # Check for the missing mapping tree.
         suffix = self.get_attr_val_utf8('nsslapd-suffix')
-        bename = self.get_attr_val_bytes('cn')
+        bename = self.lint_uid()
         try:
             mt = self._mts.get(suffix)
-            if mt.get_attr_val_bytes('nsslapd-backend') != bename and mt.get_attr_val('nsslapd-state') != ensure_bytes('backend'):
+            if mt.get_attr_val_utf8('nsslapd-backend') != bename and mt.get_attr_val_utf8('nsslapd-state') != 'backend':
                 raise ldap.NO_SUCH_OBJECT("We have a matching suffix, but not a backend or correct database name.")
         except ldap.NO_SUCH_OBJECT:
             result = DSBLE0001
+            result['check'] = f'backends:{bename}:mappingtree'
             result['items'] = [bename, ]
             yield result
-        return None
+
+    def _lint_cl_trimming(self):
+        """Check that cl trimming is at least defined to prevent unbounded growth"""
+        bename = self.lint_uid()
+        suffix = self.get_attr_val_utf8('nsslapd-suffix')
+        replicas = Replicas(self._instance)
+        try:
+            # Check if replication is enabled
+            replicas.get(suffix)
+            # Check the changelog
+            cl = Changelog(self._instance, suffix=suffix)
+            if cl.get_attr_val_utf8('nsslapd-changelogmaxentries') is None and \
+               cl.get_attr_val_utf8('nsslapd-changelogmaxage') is None:
+                report = copy.deepcopy(DSCLLE0001)
+                report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
+                report['check'] = f'backends:{bename}::cl_trimming'
+                yield report
+        except:
+            # Suffix is not replicated
+            self.log.debug(f"_lint_cl_trimming - backend ({suffix}) is not replicated")
+            pass
 
     def create_sample_entries(self, version):
         """Creates sample entries under nsslapd-suffix value
@@ -554,7 +585,7 @@ class Backend(DSLdapObject):
 
         return (dn, valid_props)
 
-    def create(self, dn=None, properties=None, basedn=DN_LDBM):
+    def create(self, dn=None, properties=None, basedn=DN_LDBM, create_mapping_tree=True):
         """Add a new backend entry, create mapping tree,
          and, if requested, sample entries
 
@@ -564,6 +595,8 @@ class Backend(DSLdapObject):
         :type properties: dict
         :param basedn: Base DN of the new entry
         :type basedn: str
+        :param create_mapping_tree: If a related mapping tree node should be created
+        :type create_mapping_tree: bool
 
         :returns: DSLdapObject of the created entry
         """
@@ -588,16 +621,19 @@ class Backend(DSLdapObject):
         super(Backend, self).create(dn, properties, basedn)
 
         # We check if the mapping tree exists in create, so do this *after*
-        properties = {
-            'cn': self._nprops_stash['nsslapd-suffix'],
-            'nsslapd-state': 'backend',
-            'nsslapd-backend': self._nprops_stash['cn'],
-        }
-        if parent_suffix:
-            # This is a subsuffix, set the parent suffix
-            properties['nsslapd-parent-suffix'] = parent_suffix
-        self._mts.create(properties=properties)
-        if sample_entries is not False:
+        if create_mapping_tree is True:
+            properties = {
+                'cn': self._nprops_stash['nsslapd-suffix'],
+                'nsslapd-state': 'backend',
+                'nsslapd-backend': self._nprops_stash['cn'],
+            }
+            if parent_suffix:
+                # This is a subsuffix, set the parent suffix
+                properties['nsslapd-parent-suffix'] = parent_suffix
+            self._mts.create(properties=properties)
+
+        # We can't create the sample entries unless a mapping tree was installed.
+        if sample_entries is not False and create_mapping_tree is True:
             self.create_sample_entries(sample_entries)
         return self
 
@@ -750,7 +786,7 @@ class Backend(DSLdapObject):
                 enc_attr.delete()
                 break
 
-    def import_ldif(self, ldifs, chunk_size=None, encrypted=False, gen_uniq_id=False, only_core=False,
+    def import_ldif(self, ldifs, chunk_size=None, encrypted=False, gen_uniq_id=None, only_core=False,
                     include_suffixes=None, exclude_suffixes=None):
         """Do an import of the suffix"""
 
@@ -848,6 +884,10 @@ class Backends(DSLdapObjects):
         self._childobject = Backend
         self._basedn = DN_LDBM
 
+    @classmethod
+    def lint_uid(cls):
+        return 'backends'
+
     def import_ldif(self, be_name, ldifs, chunk_size=None, encrypted=False, gen_uniq_id=None, only_core=False,
                     include_suffixes=None, exclude_suffixes=None):
         """Do an import of the suffix"""
@@ -931,11 +971,22 @@ class Backends(DSLdapObjects):
         task = task.create(properties=task_properties)
         return task
 
+    def delete_all_dangerous(self):
+        """
+        Delete all backends. This deletes from longest to shortest suffix
+        to ensure correct delete ordering.
+        """
+        for be in sorted(self.list(), key=lambda be: len(be.get_suffix()), reverse=True):
+            be.delete()
+
 
 class DatabaseConfig(DSLdapObject):
-    """Chaining Default Config settings DSLdapObject with:
-    - must attributes = ['cn']
-    - RDN attribute is 'cn'
+    """Backend Database configuration
+
+    The entire database configuration consists of the  main global configuration entry,
+    and the underlying DB library configuration: whither BDB or LMDB.  The combined
+    configuration should be presented as a single entity so the end user does not need
+    to worry about what library is being used, and just focus on the configuration.
 
     :param instance: An instance
     :type instance: lib389.DirSrv
@@ -943,12 +994,101 @@ class DatabaseConfig(DSLdapObject):
     :type dn: str
     """
 
-    _must_attributes = ['cn']
-
-    def __init__(self, instance, dn=None):
+    def __init__(self, instance, dn="cn=config,cn=ldbm database,cn=plugins,cn=config"):
         super(DatabaseConfig, self).__init__(instance, dn)
         self._rdn_attribute = 'cn'
         self._must_attributes = ['cn']
+        self._global_attrs = [
+            'nsslapd-lookthroughlimit',
+            'nsslapd-mode',
+            'nsslapd-idlistscanlimit',
+            'nsslapd-directory',
+            'nsslapd-import-cachesize',
+            'nsslapd-idl-switch',
+            'nsslapd-search-bypass-filter-test',
+            'nsslapd-search-use-vlv-index',
+            'nsslapd-exclude-from-export',
+            'nsslapd-serial-lock',
+            'nsslapd-subtree-rename-switch',
+            'nsslapd-pagedlookthroughlimit',
+            'nsslapd-pagedidlistscanlimit',
+            'nsslapd-rangelookthroughlimit',
+            'nsslapd-backend-opt-level',
+            'nsslapd-backend-implement',
+        ]
+        self._db_attrs = {
+            'bdb':
+                [
+                    'nsslapd-dbcachesize',
+                    'nsslapd-db-logdirectory',
+                    'nsslapd-db-home-directory',
+                    'nsslapd-db-durable-transaction',
+                    'nsslapd-db-transaction-wait',
+                    'nsslapd-db-checkpoint-interval',
+                    'nsslapd-db-compactdb-interval',
+                    'nsslapd-db-page-size',
+                    'nsslapd-db-transaction-batch-val',
+                    'nsslapd-db-transaction-batch-min-wait',
+                    'nsslapd-db-transaction-batch-max-wait',
+                    'nsslapd-db-logbuf-size',
+                    'nsslapd-db-locks',
+                    'nsslapd-db-private-import-mem',
+                    'nsslapd-import-cache-autosize',
+                    'nsslapd-cache-autosize',
+                    'nsslapd-cache-autosize-split',
+                    'nsslapd-import-cachesize',
+                    'nsslapd-search-bypass-filter-test',
+                    'nsslapd-serial-lock',
+                    'nsslapd-db-deadlock-policy',
+                ],
+            'lmdb': []
+        }
         self._create_objectclasses = ['top', 'extensibleObject']
         self._protected = True
+        # This could be "bdb" or "lmdb", use what we have configured in the global config
+        self._db_lib = self.get_attr_val_utf8_l('nsslapd-backend-implement')
         self._dn = "cn=config,cn=ldbm database,cn=plugins,cn=config"
+        self._db_dn = f"cn={self._db_lib},cn=config,cn=ldbm database,cn=plugins,cn=config"
+        self._globalObj = DSLdapObject(self._instance, dn=self._dn)
+        self._dbObj = DSLdapObject(self._instance, dn=self._db_dn)
+        # Assert there is no overlap in different config sets
+        assert_c(len(set(self._global_attrs).intersection(set(self._db_attrs['bdb']), set(self._db_attrs['lmdb']))) == 0)
+
+    def get(self):
+        """Get the combined config entries"""
+        # Get and combine both sets of attributes
+        global_attrs = self._globalObj.get_attrs_vals_utf8(self._global_attrs)
+        db_attrs = self._dbObj.get_attrs_vals_utf8(self._db_attrs[self._db_lib])
+        combined_attrs = {**global_attrs, **db_attrs}
+        return combined_attrs
+
+    def display(self):
+        """Display the combined configuration"""
+        global_attrs = self._globalObj.get_attrs_vals_utf8(self._global_attrs)
+        db_attrs = self._dbObj.get_attrs_vals_utf8(self._db_attrs[self._db_lib])
+        combined_attrs = {**global_attrs, **db_attrs}
+        for (k, vo) in combined_attrs.items():
+            if len(vo) == 0:
+                vo = ""
+            else:
+                vo = vo[0]
+            self._instance.log.info(f'{k}: {vo}')
+
+    def get_db_lib(self):
+        """Return the backend library, bdb, lmdb, etc"""
+        return self._db_lib
+
+    def set(self, value_pairs):
+        for attr, val in value_pairs:
+            attr = attr.lower()
+            if attr in self._global_attrs:
+                global_config = DSLdapObject(self._instance, dn=self._dn)
+                global_config.replace(attr, val)
+            elif attr in self._db_attrs['bdb']:
+                db_config = DSLdapObject(self._instance, dn=self._db_dn)
+                db_config.replace(attr, val)
+            elif attr in self._db_attrs['lmdb']:
+                pass
+            else:
+                # Unknown attribute
+                raise ValueError("Can not update database configuration with unknown attribute: " + attr)
